import re
import hashlib
import requests
import logging
from typing import Union
from pathlib import Path
import typer
import petl as etl
from rich import print
from requests.adapters import HTTPAdapter, Retry


logging.basicConfig(level=logging.ERROR, format='(%(levelname)s): %(message)s')
logger = logging.getLogger(__name__)


def exit_if_no_data(table: Union[etl.Table, None]) -> None:
    if table is None:
        print("No data in table. Exiting.")
        raise typer.Exit()


def parse_accession_id(filename: Path) -> Union[str, None]:
    """Get accession ID from file path

    Arguments:
        filename -- Input file path 

    Returns:
        GCST or None
    """
    gcst_search = re.search(r"GCST[0-9]+", filename.stem)
    return gcst_search.group() if gcst_search else None


def parse_genome_assembly(filename: Path) -> Union[str, None]:
    """Get genome assembly from file path

    Arguments:
        filename -- Input file path 

    Returns:
        Genome assembly or None
    """
    gcst_search = re.search(r"([Build|[GRCh]*)([0-9]+)\.", filename.stem, re.IGNORECASE)
    return gcst_search.group(2) if gcst_search else None


def download_with_requests(url, params=None, headers=None) \
    -> Union[bytes, None]:
    """Download content from URL

    Arguments:
        url -- request url
        params -- requests parameters
        headers -- request headers

    Return:
        Content from URL if status code is 200 or None
    """
    s = requests.Session()
    retries = Retry(total=5, backoff_factor=2, status_forcelist=[429, 500, 502, 503, 504])
    s.mount(url, HTTPAdapter(max_retries=retries))
    try:
        r = s.get(url, params=params, headers=headers)
        status_code = r.status_code
        if status_code == 200:
            return r.content
        else:
            logger.error(f"{url} returned {status_code} status code")
            return None
    except requests.exceptions.RequestException as e:
        logger.error(e)
        return None


def set_data_outfile_name(data_infile: Path, data_outfile: str) -> str:
    if data_outfile is None:
        accession_id = parse_accession_id(filename=data_infile)
        if accession_id:
            data_outfile = accession_id + ".tsv.gz"
        else:
            data_outfile = data_infile.stem + "-REFORMED.tsv.gz"
    return data_outfile


def set_metadata_outfile_name(data_outfile: str, metadata_outfile: str) -> str:
    if metadata_outfile is None:
        metadata_outfile = data_outfile + "-meta.yaml"
    return metadata_outfile


def get_md5sum(file: Path) -> str:
    hash_md5 = hashlib.md5()
    with open(file, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()



from pathlib import Path
from typing import Union

from gwas_sumstats_tools.interfaces.data_table import SumStatsTable
from gwas_sumstats_tools.interfaces.metadata import MetadataClient


class Formatter:
    def __init__(self) -> None:
        pass

    
def format(filename: Path,
           data_outfile: Path = None,
           minimal_to_standard: bool = False,
           generate_metadata: bool = False,
           metadata_outfile: Path = None,
           metadata_infile: Path = None,
           metadata_edit_mode: bool = False,
           metadata_from_gwas_cat: bool = False,
           custom_header_map: bool = False
           ): -> None
    
   # Set data outfile name
    ss_out = set_data_outfile_name(data_infile=filename,
                                   data_outfile=data_outfile)
    # Set metadata outfile name
    m_out = set_metadata_outfile_name(data_outfile=str(filename),
                                      metadata_outfile=metadata_outfile)
    if minimal_to_standard:
        m_out = set_metadata_outfile_name(data_outfile=ss_out,
                                          metadata_outfile=metadata_outfile)
        sst = SumStatsTable(filename)
        exit_if_no_data(table=sst.sumstats)
        print("[bold]\n-------- SUMSTATS DATA --------\n[/bold]")
        print(sst.sumstats)
        if custom_header_map:
            header_map = header_dict_from_args(args=extra_args.args)
            sst.reformat_header(header_map=header_map)
        else:
            sst.reformat_header()
        sst.normalise_missing_values()
        print("[bold]\n-------- REFORMATTED DATA --------\n[/bold]")
        print(sst.sumstats)
        print(f"[green]Formatting and writing sumstats data --> {ss_out}[/green]")
        with Progress(SpinnerColumn(finished_text="Complete!"),
                      TextColumn("[progress.description]{task.description}"),
                      transient=True
                      ) as progress:
            progress.add_task(description="Processing...", total=None)
            sst.to_file(outfile=ss_out)
    # Get metadata
    if generate_metadata:
        print("[bold]\n---------- METADATA ----------\n[/bold]")
        meta_dict = get_file_metadata(in_file=filename, out_file=ss_out)
        if metadata_from_gwas_cat:
            meta_dict = metadata_dict_from_gwas_cat(accession_id=parse_accession_id(filename=filename))
        if metadata_edit_mode:
            meta_dict.update(metadata_dict_from_args(args=extra_args.args))
        ssm = MetadataClient(out_file=m_out,
                             in_file=metadata_infile)
        ssm.update_metadata(data_dict=meta_dict)
        print(ssm)
        print(f"[green]Writing metadata --> {m_out}[/green]")
        ssm.to_file()
    if not any([minimal_to_standard, generate_metadata]):
        print("Nothing to do.")


from pathlib import Path
from typing import Union
import pandas as pd
import numpy as np
import petl as etl


"""formatters

rename_headers(chromosome=...,)

order_headers(custome_order=False)

replace_values(all=False, field, from, to)
    petl.transform.conversions.replaceall()
    petl.transform.conversions.replace()

coerce_chromosome()

normalise_missing_values()

"""


class SumStatsTable:
    FIELD_MAP = {"variant_id": "rsid"}
    FIELDS_REQUIRED = ("chromosome", "base_pair_location", "effect_allele",
                       "other_allele", "standard_error",
                       "effect_allele_frequency", "p_value")
    FIELDS_EFFECT = ("beta", "odds_ratio", "hazard_ratio")
    FIELDS_OPTIONAL = ("variant_id", "rsid", "info", "ci_upper", "ci_lower", "ref_allele")

    def __init__(self, sumstats_file: Path, delimiter: str = None) -> None:
        self.filename = str(sumstats_file)
        self.delimiter = delimiter if delimiter else self._get_delimiter(sumstats_file)
        self.sumstats = self.from_file(infile=self.filename)
        self.sumstats_df = None

    def reformat_header(self, header_map: dict = FIELD_MAP) -> etl.Table:
        """Reformats the headers according to the standard

        Returns:
            etl.Table
        """
        self.rename_headers(header_map=header_map)
        missing_headers = self._get_missing_headers()
        if missing_headers:
            self._add_missing_headers(missing_headers)
        header_order = self._set_header_order()
        self.sumstats = etl.cut(self.sumstats, *header_order)
        return self.sumstats

    def from_file(self, infile: str) -> Union[etl.Table, None]:
        """Try to read the file in to a Table.
        Files can be TAB seperated and optionally compressed
        with (B)GZIP. There could be cases where an input file 
        has been renamed but the data is something different 
        to that suggested by the name and extension. Most 
        cases should be covered by the exception clause.

        Arguments:
            infile -- Input file

        Returns:
            petl Table or None
        """
        try:
            self.sumstats = etl.fromcsv(self.filename, delimiter=self.delimiter)
            if etl.nrows(self.head_table(nrows=1)) < 1:
                return None
            return self.sumstats
        except (IOError, UnicodeDecodeError) as exception:
            print(exception)
            return None

    def to_file(self, outfile: Path) -> None:
        """Write table to TSV file

        Arguments:
            outfile -- Output file name
        """
        self.sumstats.totsv(str(outfile))

    def head_table(self, nrows: int = 10) -> etl.Table:
        return etl.head(self.sumstats, n=nrows)

    def _get_delimiter(self, filepath: Path) -> str:
        """Get delimiter from file path

        Arguments:
            filepath -- Input file path

        Returns:
            delimiter, either ',' or '\t'
        """
        if not isinstance(filepath, Path):
            filepath = Path(filepath)
        return ',' if '.csv' in filepath.suffixes else '\t'

    def rename_headers(self, header_map: dict) -> etl.Table:
        """Rename headers according to the header map

        Keyword Arguments:
            header_map -- dict of header mappings from old to new \
                (default: {HEADER_MAP})

        Returns:
            etl.Table
        """
        filtered_header_map = {k: v for k, v in header_map.items() if k in self.header()}
        self.sumstats = etl.rename(self.sumstats, filtered_header_map)
        return self.sumstats

    def normalise_missing_values(self) -> etl.Table:
        self.sumstats = etl.replaceall(self.sumstats, 'NA', '#NA')
        self.sumstats = etl.replaceall(self.sumstats, None, '#NA')
        self.sumstats = etl.replaceall(self.sumstats, '', '#NA')
        return self.sumstats

    def _get_missing_headers(self) -> set:
        """Identify and return missing headers

        Returns:
            set of missing headers
        """
        missing_headers = set(self.FIELDS_REQUIRED) - set(self.header())
        if set(self.FIELDS_EFFECT).isdisjoint(set(self.header())):
            missing_headers.add("beta")
        return missing_headers

    def _set_header_order(self) -> list:
        """Set the header order

        Returns:
            List of headers in standard order
        """
        all_headers = [h for h in self.FIELDS_REQUIRED]
        all_headers.extend([h for h in self.FIELDS_OPTIONAL])
        all_headers.extend([h for h in self.FIELDS_EFFECT])
        header_order = [h for h in self.FIELDS_REQUIRED]
        header_order.extend([h for h in self.FIELDS_OPTIONAL if h in self.header()])
        header_order.extend([h for h in self.header() if h not in all_headers])
        if 'beta' in self.header():
            header_order.insert(4, 'beta')
            for h in ('odds_ratio', 'hazard_ratio'):
                header_order.append(h) if h in self.header() else None
        elif 'beta' not in self.header() and 'odds_ratio' in self.header():
            header_order.insert(4, 'odds_ratio')
            header_order.append('hazard_ratio') if 'hazard_ratio' in self.header() else None
        elif 'odds_ratio' not in self.header() and 'hazard_ratio' in self.header():
            header_order.insert(4, 'hazard_ratio')
        else:
            pass
        return header_order

    def _add_missing_headers(self, missing_headers: set) -> etl.Table:
        """Add missing headers and fill with #NA

        Arguments:
            missing_headers -- missing headers set

        Returns:
            etl.Table
        """
        add_fields = [(h, '#NA') for h in missing_headers]
        if len(add_fields) > 1:
            self.sumstats = etl.addfields(self.sumstats, add_fields)
        else:
            field, value = add_fields[0]
            self.sumstats = etl.addfield(self.sumstats, field, value)
        return self.sumstats

    def header(self) -> tuple:
        """Get the header of the file

        Returns:
            tuple of the headers
        """
        if self.sumstats:
            return etl.header(self.sumstats)
        return ()

    def effect_field(self) -> Union[str, None]:
        """Get the effect allele (field index 4).

        Returns:
            effect field label
        """
        field_4 = self._get_field_label_from_index(4)
        return field_4
    
    def p_value_field(self) -> Union[str, None]:
        """Get the p_value field (field index 7).

        Returns:
            p_value field label
        """
        field_4 = self._get_field_label_from_index(7)
        return field_4

    def _get_field_label_from_index(self, index: int) -> Union[str, None]:
        """Get the field label from specified index

        Arguments:
            index -- index of field

        Returns:
            field label or None
        """
        return self.header()[index] if len(self.header()) > index else None

    def _pval_to_mantissa_and_exponent(self, table: etl.Table) -> etl.Table:
        table_w_split_p = etl.split(self._square_up_table(table),
                                    self.p_value_field(),
                                    'e|E',
                                    newfields=['_p_value_mantissa',
                                               '_p_value_exponent'],
                                    include_original=True,
                                    maxsplit=1)
        return table_w_split_p

    def _prep_table_for_validation(self) -> etl.Table:
        table = etl.Table()
        if self.sumstats:
            table = self._pval_to_mantissa_and_exponent(table=self.sumstats)
        return table

    def as_pd_df(self, nrows: int = None) -> pd.DataFrame:
        """Sumstats table as a Pandas dataframe

        Keyword Arguments:
            nrows -- Number of rows (default: {None, which means all rows})

        Returns:
            Pandas dataframe
        """

        df = pd.DataFrame()
        if self.sumstats:
            table_to_validate = self._square_up_table(self._prep_table_for_validation())
            df = etl.todataframe(table_to_validate, nrows=nrows)
            df = df.replace([None, "", "#NA", "NA", "N/A", "NaN"], np.nan)
        return df
    
    def _square_up_table(self, table: etl.Table, missing: str="#NA") -> etl.Table:
        """Square up a table with missing/extra values on rows.

        Returns:
            etl.Table
        """
        return etl.cat(table, missing=missing)


def header_dict_from_args(args: list) -> dict:
    """Generate a dict from cli args split on ":"

    Arguments:
        args -- cli args list

    Returns:
        Dict of key, values
    """
    header_dict = {}
    for arg in args:
        if ":" not in arg:
            # skip because it's not a metadata mapping
            pass
        else:
            key, value = arg.replace("--", "").split(":")
            header_dict[key] = value
    return header_dict




import yaml
from typing import Union
from datetime import date
from pathlib import Path
from gwas_sumstats_tools.utils import (download_with_requests,
                                       parse_accession_id,
                                       parse_genome_assembly,
                                       get_md5sum)
from gwas_sumstats_tools.schema.metadata import SumStatsMetadata


GWAS_CAT_API_STUDIES_URL = "https://www.ebi.ac.uk/gwas/rest/api/studies/"
GWAS_CAT_MAPPINGS = {
    'genotypingTechnology': 'genotyping_technology',
    'sampleSize': 'sample_size',
    'sampleAncestry': 'sample_ancestry',
    'traitDescription': 'trait_description',
    'minorAlleleFreqLowerLimit': 'minor_allele_freq_lower_limit',
    'ancestryMethod': 'ancestry_method',
    'caseControlStudy': 'case_control_study',
    'caseCount': 'case_count',
    'controlCount': 'control_count',
    'genomeAssembly': 'genome_assembly',
    'analysisSoftware': 'analysis_software',
    'imputationPanel': 'imputation_panel',
    'imputationSoftware': 'imputation_software',
    'adjustedCovariates': 'adjusted_covariates',
    'ontologyMapping': 'ontology_mapping',
    'authorNotes': 'author_notes',
    'coordinateSystem': 'coordinate_system',
    'sex': 'sex'
    }
GENOME_ASSEMBLY_MAPPINGS = {
    '36': 'GRCh36',
    '37': 'GRCh37',
    '38': 'GRCh38'
    }


class MetadataClient:
    def __init__(self, meta_dict: dict = {},
                 in_file: Path = None,
                 out_file: Path = None) -> None:
        """SumStats Metadata client

        Keyword Arguments:
            meta_dict -- Dict of metadata (default: {None})
            in_file -- Input metadata YAML file (default: {None})
            out_file -- Output metadata YAML file (default: {None})
        """
        self.metadata = SumStatsMetadata.construct()
        self._meta_dict = meta_dict
        self._in_file = in_file
        self._out_file = out_file

    def from_file(self) -> None:
        """Create metadata from YAML file
        """
        with open(self._in_file, "r") as fh:
            self._meta_dict = yaml.safe_load(fh)
            self.update_metadata(self._meta_dict)

    def to_file(self) -> None:
        """Write metadata to YAML file
        """
        with open(self._out_file, "w") as fh:
            yaml.dump(self.metadata.dict(exclude_none=True),
                      fh,
                      encoding='utf-8')

    def update_metadata(self, data_dict: dict) -> None:
        """Create a copy of the model and update (no validation occurs)

        Arguments:
            data_dict -- Dict of data to populate model
        """
        self._meta_dict.update(data_dict)
        self.metadata = self.metadata.parse_obj(self._meta_dict)

    def __repr__(self) -> str:
        """Representation of metadata.
        """
        return self.as_yaml()

    def as_dict(self) -> dict:
        """Dict repr of metadata

        Returns:
            Dict of metadata
        """
        return self.metadata.dict()

    def as_yaml(self, **kwargs) -> str:
        return yaml.dump(self.metadata.dict(**kwargs))
        


def metadata_dict_from_args(args: list) -> dict:
    """Generate a dict from cli args split on "="

    Arguments:
        args -- cli args list

    Returns:
        Dict of key, values
    """
    meta_dict = {}
    for arg in args:
        if "=" not in arg:
            # skip because it's not a metadata mapping
            pass
        else:
            key, value = arg.replace("--", "").split("=")
            meta_dict[key] = value
    return meta_dict


def metadata_dict_from_gwas_cat(accession_id: str) -> dict:
    """Extract metadat from the GWAS Catalog API

    Arguments:
        accession_id -- GWAS Catalog accession ID

    Returns:
        Metadata dict
    """
    study_url = GWAS_CAT_API_STUDIES_URL + accession_id
    content = download_with_requests(url=study_url)
    meta_dict = {}
    # TODO parse content into meta_dict
    return meta_dict


def get_file_metadata(in_file: Path, out_file: str) -> dict:
    """Get file related metadata

    Arguments:
        in_file -- sumstats in file
        outfile -- sumstats out file

    Returns:
        Metadata dict
    """
    meta_dict = {}
    accession_id = parse_accession_id(filename=in_file) 
    meta_dict['gwas_id'] = accession_id
    meta_dict['data_file_name'] = Path(out_file).name
    meta_dict['file_type'] = 'GWAS-SFF v1.0'
    meta_dict['genome_assembly'] = GENOME_ASSEMBLY_MAPPINGS.get(parse_genome_assembly(filename=in_file), 'unknown')
    meta_dict['data_file_md5sum'] = get_md5sum(out_file) if Path(out_file).exists() else None
    meta_dict['date_last_modified'] = date.today()
    meta_dict['gwas_catalog_api'] = GWAS_CAT_API_STUDIES_URL + accession_id
    return meta_dict


def init_metadata_from_file(filename: Path, metadata_infile: Path = None) -> Union[SumStatsMetadata, None]:
    m_in = metadata_infile if metadata_infile else filename.with_suffix(filename.suffix + "-meta.yaml")
    if m_in.exists():
        ssm = MetadataClient(in_file=m_in)
        ssm.from_file()
        return ssm
    else:
        return None


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gwas_sumstats_tools',
 'gwas_sumstats_tools.interfaces',
 'gwas_sumstats_tools.schema']

package_data = \
{'': ['*']}

install_requires = \
['hypothesis>=6.68.2,<7.0.0',
 'pandera[io]>=0.13.4,<0.14.0',
 'petl>=1.7.12,<2.0.0',
 'pydantic>=1.10.4,<2.0.0',
 'pytest-mock>=3.10.0,<4.0.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.28.2,<3.0.0',
 'typer[all]>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['gwas-ssf = gwas_sumstats_tools.cli:app']}

setup_kwargs = {
    'name': 'gwas-sumstats-tools',
    'version': '0.1.3a0',
    'description': '',
    'long_description': '# GWAS SumStats Tools\n\n\nA basic toolkit for reading and formatting GWAS sumstats files from the GWAS Catalog.\nBuilt with:\n* [Petl](https://petl.readthedocs.io/en/stable/index.html)\n* [Pydantic](https://docs.pydantic.dev/)\n* [Typer](https://typer.tiangolo.com/)\n\nThere are three commands, `read`, `validate` and `format`.\n\n`read` is for:\n* Previewing a data file: _no options_\n* Extracting the field headers: `-h`\n* Extracting all the metadata: `-M`\n* Extacting specific field, value pairs from the metada: `-m <field name>`\n\n`validate` is for:\n* Validating a summary statistic file using a dynamically generated schema\n\n`format` is for:\n* Converting a minamally formatted sumstats data file to the standard format. This is not guaranteed to return a valid standard file, because manadatory data fields could be missing in the input. It simply does the following. `-s`\n  * Renames `variant_id` -> `rsid`\n  * Reorders the fields\n  * Converts `NA` missing values to `#NA`\n  * It is memory efficient and will take approx. 30s per 1 million records\n* Generate metadata for a data file: `-m`\n  * Read metadata in from existing file: `--meta-in <file>`\n  * Create metadata from the GWAS Catalog (internal use, requires authenticated API): `-g`\n  * Edit/add the values to the metadata: `-e` with `--<FIELD>=<VALUE>`\n\n## Installation\n```console\n$ pip install gwas-sumstats-tools\n```\n\n## Usage\n\n```console\n$ gwas-ssf [OPTIONS] COMMAND [ARGS]...\n```\n\n**Options**:\n\n* `--help`: Show this message and exit.\n\n**Commands**:\n\n* `format`: Format a sumstats file and...\n* `read`: Read a sumstats file\n\n### `gwas-ssf read`\n\nRead (preview) a sumstats file\n\n**Usage**:\n\n```console\n$ gwas-ssf read [OPTIONS] FILENAME\n```\n\n**Arguments**:\n\n* `FILENAME`: Input sumstats file  [required]\n\n**Options**:\n\n* `-h, --get-header`: Just return the headers of the file  [default: False]\n* `--meta-in PATH`: Specify a metadata file to read in, defaulting to <filename>-meta.yaml\n* `-M, --get-all-metadata`: Return all metadata  [default: False]\n* `-m, --get-metadata TEXT`: Get metadata for the specified fields e.g. `-m genomeAssembly -m isHarmonised\n* `--help`: Show this message and exit.\n\n\n### `gwas-ssf validate`\n\nValidate a sumstats file\n\n**Usage**:\n\n```console\n$ gwas-ssf validate [OPTIONS] FILENAME\n```\n\n**Arguments**:\n\n* `FILENAME`: Input sumstats file. Must be TSV or CSV and may be gzipped [required]\n\n**Options**:\n\n* `-e, --errors-out`: Output erros to a csv file, <filename>.err.csv.gz\n* `-z, --p-zero`: Force p-values of zero to be allowable. Takes precedence over inferred value (-i)\n* `-n, --p-neg-log`: Force p-values to be validated as -log10. Takes precedence over inferred value (-i)\n* `-m, --min-rows`:  Minimum rows acceptable for the file [default: 100000]\n* `-i, --infer-from-metadata`: Infer validation options from the metadata file <filename>-meta.yaml. E.g. fields for analysis software and negative log10 p-values affect the data validation behaviour.\n* `--help`: Show this message and exit.\n\n### `gwas-ssf format`\n\nFormat a sumstats file and creating a new one. Add/edit metadata.\n\n**Usage**:\n\n```console\n$ gwas-ssf format [OPTIONS] FILENAME\n```\n\n**Arguments**:\n\n* `FILENAME`: Input sumstats file. Must be TSV or CSV and may be gzipped  [required]\n\n**Options**:\n\n* `-o, --ss-out PATH`: Output sumstats file\n* `-s, --minimal2standard`: Try to convert a valid, minimally formatted file to the standard format.This assumes the file at least has `p_value`  combined with rsid in `variant_id` field or `chromosome` and `base_pair_location`. Validity of the new file is not guaranteed because mandatory data could be missing from the original file.  [default: False]\n* `-m, --generate-metadata`: Create the metadata file  [default: False]\n* `--meta-out PATH`: Specify the metadata output file\n* `--meta-in PATH`: Specify a metadata file to read in\n* `-e, --meta-edit`: Enable metadata edit mode. Then provide params to edit in the `--<FIELD>=<VALUE>` format e.g. `--GWASID=GCST123456` to edit/add that value  [default: False]\n* `-g, --meta-gwas`: Populate metadata from GWAS Catalog  [default: False]\n* `-c, --custom-header-map`: Provide a custom header mapping using the `--<FROM>:<TO>` format e.g. `--chr:chromosome`  [default: False]\n* `--help`: Show this message and exit.\n\n## Development\nThis repository uses [poetry](https://python-poetry.org/docs/) for dependency and packaging management.\n\nTo run the tests:\n\n1. [install poetry](https://python-poetry.org/docs/#installation)\n\n2. `git clone https://github.com/EBISPOT/gwas-sumstats-tools.git`\n3. `cd gwas-sumstats-tools`\n4. `poetry install`\n5. `poetry run pytest`\n',
    'author': 'jdhayhurst',
    'author_email': 'jhayhurst@ebi.ac.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
